-----------------------------------------------------------------------------------------------
-- Client Lua Script for SetUsUpTheBomb 
-- @author Kihra, but borrows heavily from daihenka's Threat Meter add-on code
-----------------------------------------------------------------------------------------------
 
require "Apollo"
require "Window"
require "GameLib"
require "GroupLib"
require "Sound"
require "MatchingGame"
 

-----------------------------------------------------------------------------------------------
-- SetUsUpTheBomb Module Definitions
-----------------------------------------------------------------------------------------------
local ADDON_NAME     = "SetUsUpTheBomb"
local CONFIG_VERSION = 1

local kBombName = "Detonation Bomb"
local kBossName = "Phage Maw"
local kCountdownTime = 90

local SetUsUpTheBomb = Apollo.GetAddon(ADDON_NAME)
local DaiUtil     = Apollo.GetPackage("DaiUtil-1.0").tPackage

-----------------------------------------------------------------------------------------------
-- Constants and Local Variables
-----------------------------------------------------------------------------------------------
local setmetatable, select, unpack, pairs, ipairs = setmetatable, select, unpack, pairs, ipairs
local ostime = os.time
local tremove, tinsert, tsort = table.remove, table.insert, table.sort
local strformat = string.format

-- Convert large numbers into compact variations
-- @param nArg the value to compact
-- @return (string) The compact version of the number
local function FormatBigNumber(nArg, nPrecision)
  if type(nArg) == "string" then
    nArg = tonumber(nArg)
  end
  
	nPrecision = nPrecision or 1
	
	-- Turns 99999 into 99k and 90000 into 90k
	if nArg < 1000 then
		return nArg
	elseif nArg < 1000000 then
		return string.format("%."..nPrecision.."fk", nArg/1000)
	elseif nArg < 1000000000 then
		return string.format("%."..nPrecision.."fm", nArg/1000000)
	else
		return string.format("%."..nPrecision.."fg", nArg/1000000000)
	end
end


-----------------------------------------------------------------------------------------------
-- Core Addon Functions
-----------------------------------------------------------------------------------------------
function SetUsUpTheBomb:OnInitialize()
  -- slash commands
  Apollo.RegisterSlashCommand("bomb",      "OnSlashCmd", self)

  -- Register handlers for events
  Apollo.RegisterEventHandler("UnitEnteredCombat",        "OnUnitEnteredCombat",        self)
 
 -- Unit Discovery
  Apollo.RegisterEventHandler("UnitCreated", 						"OnUnitCreated", 						self)
  Apollo.RegisterEventHandler("UnitDestroyed", 					"OnUnitDestroyed",						self)
  Apollo.RegisterEventHandler("CombatLogDamage", 					"OnCombatLogDamage",					self)
		
  Apollo.RegisterTimerHandler("BombUpdateTimer", "OnUnitUpdate", self)
Apollo.RegisterTimerHandler("BossUpdateTimer", "OnBossTimerUpdate", self)


  self.tItems                = {}
  self.tDisplayOrder         = {}
  self.tMarkerOrder = {}
  self.bInCombat             = false
  self.tUnits  = {}
  self.nUnitCount = 0

  self.lastBossMooTime = 0

  self.wndMain    = self:CreateMainWindow()
  self.wndWarning = self:CreateWarningWindow()

  self.wndItemList = self.wndMain:FindChild("ItemList")
end

function SetUsUpTheBomb:OnEnable()
  self.bInCombat = GameLib.GetPlayerUnit():IsInCombat()
end

-----------------------------------------------------------------------------------------------
-- SetUsUpTheBomb Settings
-----------------------------------------------------------------------------------------------
--- Addon save configuration event handler
-- @param eLevel addon save level
-- @return table with configuration data to save
function SetUsUpTheBomb:OnSaveSettings(eLevel)
  if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
    return
  end
	
  local tData = DaiUtil.TableCopy(self.db)
  tData.Version = CONFIG_VERSION
  tData.tAnchorOffsets = {self.wndMain:GetAnchorOffsets()}
  
  DaiUtil.SerializeColors(tData)
  
  return tData
end

--- Addon restore configuration event handler
-- @param eLevel addon save level
-- @param tData saved addon configuration data
function SetUsUpTheBomb:OnRestoreSettings(eLevel, tData)
  if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
    return
  end
  
  if (tData.Version or 0) < CONFIG_VERSION then
    -- discard old versions of the config as they are incompatible
    return
  end
  
  DaiUtil.DeserializeColors(tData)

  if type(tData.tAnchorOffsets) == "table" and #tData.tAnchorOffsets == 4 then
    self.wndMain:SetAnchorOffsets(unpack(tData.tAnchorOffsets))
    tData.tAnchorOffsets = nil
  end

  -- merge the settings with the defaults
  DaiUtil.TableMerge(self.db, tData)
end


-----------------------------------------------------------------------------------------------
-- SetUsUpTheBomb Slash Commands
-----------------------------------------------------------------------------------------------
--- General slash command handler
-- @param cmd
-- @param args
function SetUsUpTheBomb:OnSlashCmd(cmd, args)
  self.wndMain:Show(true) -- show the window
end


-----------------------------------------------------------------------------------------------
-- SetUsUpTheBomb Visibility
-----------------------------------------------------------------------------------------------

function SetUsUpTheBomb:OnClose( wndHandler, wndControl, eMouseButton )
	self.wndMain:Show(false)
end

function SetUsUpTheBomb:UpdateVisibility()
  local bShow = self:ShouldShow()
  self.wndMain:Show(bShow)
end

function SetUsUpTheBomb:ShouldShow()
  if self.db.bHideWhenNotInCombat and not self.bInCombat then 
    return false 
  end  
  return self.nUnitCount > 0 or self.lastBossMooTime > 0
end

-----------------------------------------------------------------------------------------------
-- SetUsUpTheBomb Event Handlers
-----------------------------------------------------------------------------------------------

function SetUsUpTheBomb:OnUnitCreated(unit)
	local color = self.db.crNotPlayer
	if unit:GetName() ~= kBombName then
		return
	end
	
	local tUnitData = {
		unit = unit,
		id = unit:GetId(),
		name = unit:GetName(),
		type = unitType,
		cr = self.db.crNotPlayer,
		marker = 0,
		
		-- Vitals
		health = {
			current = unit:GetHealth(),
			max = math.max(1, unit:GetMaxHealth()),
		},
		shield = {
			current = unit:GetShieldCapacity(),
			max = unit:GetShieldCapacityMax(),
		},
		absorb = {
			current = 0,
			max = unit:GetAbsorptionMax() or 0,
		}
	}

	if tUnitData.absorb.max then
		tUnitData.absorb.current = unit:GetAbsorptionValue()
	end
	
	local alreadyPresent = self.tUnits[tUnitData.id] ~= nil
	if alreadyPresent then
		tUnitData.marker = self.tUnits[tUnitData.id].marker
	else
		self.nUnitCount = self.nUnitCount + 1
		if self.nUnitCount == 1 then
			Apollo.CreateTimer("BombUpdateTimer", 0.1, true)
		end
	end

	self.tUnits[tUnitData.id] = tUnitData
		
	self:UpdateVisibility()
	self:UpdateDisplay()
end

function SetUsUpTheBomb:OnUnitDestroyed(unit)
	local unitId = unit:GetId()
	if not self.tUnits[unitId] then
		return nil
	end
	
	self.tUnits[unitId] = nil
	self.nUnitCount = self.nUnitCount - 1
	
    if self.nUnitCount == 0 then
		Apollo.StopTimer("BombUpdateTimer")
	end
	
	self:UpdateVisibility()
	self:UpdateDisplay()
end

function SetUsUpTheBomb:MarkerToSprite(nMarker)
	if nMarker == 1 then
		return "Icon_Windows_UI_CRB_Marker_Bomb"
	elseif nMarker == 2 then
		return "Icon_Windows_UI_CRB_Marker_Ghost"
	elseif nMarker == 3 then
		return "Icon_Windows_UI_CRB_Marker_Mask"
	elseif nMarker == 4 then
		return  "Icon_Windows_UI_CRB_Marker_Octopus"
	elseif nMarker == 5 then
		return "Icon_Windows_UI_CRB_Marker_Pig"
	elseif nMarker == 6 then
		return "Icon_Windows_UI_CRB_Marker_Chicken"
	elseif nMarker == 7 then
		return "Icon_Windows_UI_CRB_Marker_Toaster"
	elseif nMarker == 8 then
		return  "Icon_Windows_UI_CRB_Marker_UFO"
	else
		return "Icon_Windows_UI_CRB_Marker_Bomb"
	end
end

function SetUsUpTheBomb:OnCombatLogDamage(tEventArgs)
	if tEventArgs.bTargetVulnerable and tEventArgs.unitTarget and tEventArgs.unitTarget:GetName() == kBossName then
		local oldMooTime = self.lastBossMooTime
		self.lastBossMooTime = GameLib:GetGameTime()
		if oldMooTime == 0 then
			Apollo.CreateTimer("BossUpdateTimer", 0.5, true)
			self.tUnits = {}
			self:UpdateVisibility()
			self:UpdateDisplay()
		end
	elseif tEventArgs.unitCaster and tEventArgs.unitTarget and tEventArgs.unitTarget:GetName() == kBombName and tEventArgs.unitCaster:GetClassId() <= 7 then
		local targetMarker = tEventArgs.unitCaster:GetTargetMarker() or 0
		local unitInfo = self.tUnits[tEventArgs.unitTarget:GetId()]
		if targetMarker > 0 and unitInfo and unitInfo.marker ~= targetMarker then
			for _, UnitData in pairs(self.tUnits) do
				if UnitData.marker == targetMarker then
					UnitData.marker = 0
				end
			end
			unitInfo.marker = targetMarker
			self:UpdateDisplay()
		end
	end
end

function SetUsUpTheBomb:OnUnitUpdate()
	local invalidAndDeadUnits = {}
	for _, UnitData in pairs(self.tUnits) do
		self:UpdateVitals(UnitData)
		if UnitData.unit:IsDead() or not UnitData.unit:IsValid() then
			table.insert(invalidAndDeadUnits, UnitData)
		end
	end
	for _, UnitData in pairs(invalidAndDeadUnits) do
		self:OnUnitDestroyed(UnitData.unit)
	end
	self:UpdateDisplay()
end

function SetUsUpTheBomb:OnBossTimerUpdate()
	local timeRemaining = kCountdownTime - (GameLib.GetGameTime() - self.lastBossMooTime)
	if timeRemaining <= 0 then
		Apollo.StopTimer("BossUpdateTimer")
		self.lastBossMooTime = 0
		self.wndMain:FindChild("TitleText"):SetText("Set Us Up The Bomb")
		self.wndWarning:Show(false)
		return
	end
	local displayTime = timeRemaining
	local timerString = "Set Us Up The Bomb - 1:"
	if (timeRemaining < 60) then
		timerString = "Set Us Up The Bomb - 0:"
	else
		displayTime = timeRemaining - 60
	end

	if (displayTime < 10) then
		timerString = timerString.."0"
	end
	timerString = timerString..math.floor(displayTime)
	
	self.wndMain:FindChild("TitleText"):SetText(timerString)
	
	self.wndWarning:Show(timeRemaining < 6)
end

function SetUsUpTheBomb:UpdateVitals(tUnitData)
	self:RefreshHealth(tUnitData)
	self:RefreshShield(tUnitData)
	self:RefreshAbsorb(tUnitData)
end

function SetUsUpTheBomb:RefreshHealth(UnitData)
	local current = UnitData.unit:GetHealth()

	if UnitData.health.current ~= current then
		UnitData.health.current = current
		UnitData.health.max = UnitData.unit:GetMaxHealth()
		return UnitData.health
	end
end

function SetUsUpTheBomb:RefreshShield(UnitData)
	local current = UnitData.unit:GetShieldCapacity()

	if UnitData.shield.current ~= current then
		UnitData.shield.current = current
		UnitData.shield.max = UnitData.unit:GetShieldCapacityMax()
		return UnitData.shield
	end
end

function SetUsUpTheBomb:RefreshAbsorb(UnitData)
	local current = 0
	local max = UnitData.unit:GetAbsorptionMax()
	
	if max and max > 0 and max ~= UnitData.absorb.max then
		local current_new = UnitData.unit:GetAbsorptionValue()
		if UnitData.absorb.current ~= current_new then
			current = current_new
		end
		
		if max ~= UnitData.absorb.max then
			UnitData.absorb.max = max
		end
		UnitData.absorb.current = current
		UnitData.absorb.max = max
		return UnitData.absorb
	end	
end

function SetUsUpTheBomb:OnUnitEnteredCombat(unit, bInCombat)
	if unit and unit:IsThePlayer() then
        self.bInCombat = bInCombat
	    self:UpdateVisibility()
		if not bInCombat then
			Apollo.StopTimer("BossUpdateTimer")
			self.lastBossMooTime = 0
			self.wndWarning:Show(false)
		end
	end
end

function SetUsUpTheBomb:UpdateDisplay()
	if self.wndMain:IsVisible() then
		self:ResetDisplayOrder()
		self:DisplayList()
	end
end

function SetUsUpTheBomb:CompareDisplay(index, text, marker)
	local changed = false
	if not self.tDisplayOrder[index] or (self.tDisplayOrder[index] and self.tDisplayOrder[index] ~= text) then
		self.tDisplayOrder[index] = text
		changed = true
	end
	
	if not self.tMarkerOrder[index] or (self.tMarkerOrder[index] and self.tMarkerOrder[index] ~= marker) then
		self.tMarkerOrder[index] = marker
		changed = true
	end

	return changed
end

function SetUsUpTheBomb:GetKeysSortedByValue()
	local newTable = {}
	for key, value in pairs(self.tUnits) do
		if value then -- FIXME: Should not be possible and yet here we are.
			local item = {}
			item["id"] = key
			item["health"] = value.health.current / value.health.max
			table.insert(newTable, item)
		end
	end

  	table.sort(newTable, function(a, b)
		return a["health"] < b["health"]
  	end)

  	return newTable
end

function SetUsUpTheBomb:DisplayList()
	local sortedKeys = self:GetKeysSortedByValue()
	local bArrange = false
	for k, item in ipairs(sortedKeys) do
		if not self.tItems[k] then
			self:AddItem(k)
		end
		local wnd = self.tItems[k]
		local tUnitData = self.tUnits[item["id"]]
		
		if self:CompareDisplay(k, tUnitData.name, tUnitData.marker) then
			wnd.nameText:SetText(tUnitData.name)
			wnd.progressBar:SetBarColor(tUnitData.cr)
			wnd.progressIcon:SetSprite(self:MarkerToSprite(tUnitData.marker))
			bArrange = true
		end

		wnd.amountText:SetText(strformat("%s (%.0f%%)", FormatBigNumber(tUnitData.health.current, 1), (tUnitData.health.current / tUnitData.health.max) * 100))
		wnd.progressBar:SetProgress(tUnitData.health.current / tUnitData.health.max)		
	end
	
	-- trim
	if #self.tItems > self.nUnitCount then
		for i=self.nUnitCount+1, #self.tItems do
			self.tItems[i].wnd:Destroy()
			self.tItems[i] = nil
		end
	end
	
	-- rearrange
	if bArrange then
		self.wndItemList:ArrangeChildrenVert()
	end
end

function SetUsUpTheBomb:ResetDisplayOrder()
	self.tDisplayOrder = {}
	self.tMarkerOrder = {}
end


function SetUsUpTheBomb:AddItem(i)
  -- load the window item for the list item
  local wnd = self:CreateItemWindow()

  -- keep track of the window item created
  local tItem = {}
  tItem.wnd          = wnd
  tItem.progressBar  = wnd:FindChild("ProgressBar")
  tItem.progressIcon = wnd:FindChild("ProgressIcon")
  tItem.nameText     = wnd:FindChild("NameText")
  tItem.amountText   = wnd:FindChild("AmountText")

  tItem.progressBar:SetMax(1)
  tItem.progressBar:SetProgress(0)
  tItem.nameText:SetTextColor(self.db.crNormalText)

  self.tItems[i] = tItem

  wnd:SetData(i)

  return tItem
end

function SetUsUpTheBomb:DestroyItemList()
	for idx,wnd in ipairs(self.tItems) do
		wnd:Destroy()
	end
	
	self.tItems = {}
end

function SetUsUpTheBomb:PostOnChannel(strResult)
	ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Command, strResult, "")
end

